///////////////////////////////////////////////////////////////////////////////
//  Filename: $RCSfile: MolInstance.java,v $
//  Purpose:  Molecule Weka instance.
//  Language: Java
//  Compiler: JDK 1.4
//  Authors:  Joerg K. Wegner
//  Version:  $Revision: 1.2 $
//            $Date: 2004/07/25 20:42:57 $
//            $Author: wegner $
//
//  Copyright (c) Dept. Computer Architecture, University of Tuebingen, Germany
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation version 2 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
///////////////////////////////////////////////////////////////////////////////
package joelib.algo.datamining.weka;

import weka.core.Instance;

import org.apache.log4j.Category;

import joelib.io.IOType;
import joelib.io.IOTypeHolder;

import joelib.molecule.JOEAtom;
import joelib.molecule.JOEMol;

import joelib.util.iterator.AtomIterator;


/**
 * Molecule Weka instance.
 *
 * @author    wegnerj
 * @license    GPL
 * @cvsversion    $Revision: 1.2 $, $Date: 2004/07/25 20:42:57 $
 */
public class MolInstance extends Instance
{
    //~ Static fields/initializers /////////////////////////////////////////////

    /**
     * Obtain a suitable logger.
     */
    private static Category logger = Category.getInstance(
            "joelib.algo.datamining.weka.MolInstance");
    protected static final IOType smiles = IOTypeHolder.instance().getIOType("SMILES");

    //~ Instance fields ////////////////////////////////////////////////////////

    /** The instance identifier. */
    protected JOEMol mol;

    //~ Constructors ///////////////////////////////////////////////////////////

    /**
     * Constructor that copies the attribute values and the weight from
     * the given instance. Reference to the dataset is set to null.
     * (ie. the instance doesn't have access to information about the
     * attribute types)
     *
     * @param instance the instance from which the attribute
     * values and the weight are to be copied
     */
    public MolInstance(MolInstance instance)
    {
        m_AttValues = instance.m_AttValues;
        m_Weight = instance.m_Weight;
        m_Dataset = null;
        mol = instance.mol;
    }

    /**
     * Constructor that inititalizes instance variable with given
     * values. Reference to the dataset is set to null. (ie. the instance
     * doesn't have access to information about the attribute types)
     *
     * @param _identifier the instance's identifier
     * @param weight the instance's weight
     * @param attValues a vector of attribute values
     */
    public MolInstance(JOEMol _mol, double weight, double[] attValues)
    {
        m_AttValues = attValues;
        m_Weight = weight;
        m_Dataset = null;
        mol = _mol;
    }

    /**
     * Constructor of an instance that sets weight to one, all values to
     * be missing, and the reference to the dataset to null. (ie. the instance
     * doesn't have access to information about the attribute types)
     *
     * @param numAttributes the size of the instance
     */
    public MolInstance(int numAttributes)
    {
        m_AttValues = new double[numAttributes];

        for (int i = 0; i < m_AttValues.length; i++)
        {
            m_AttValues[i] = MISSING_VALUE;
        }

        m_Weight = 1;
        m_Dataset = null;
        mol = null;
    }

    /*-------------------------------------------------------------------------*
     * constructor
     *-------------------------------------------------------------------------   */
    protected MolInstance()
    {
    }

    //~ Methods ////////////////////////////////////////////////////////////////

    /**
     * Sets molecule.
     */
    public void setMolecule(JOEMol _mol)
    {
        mol = _mol;
    }

    /*-------------------------------------------------------------------------*
     * public methods
     *-------------------------------------------------------------------------   */

    /**
     * Returns molecule.
     */
    public JOEMol getMolecule()
    {
        return mol;
    }

    /**
     * Produces a shallow copy of this instance. The copy has
     * access to the same dataset. (if you want to make a copy
     * that doesn't have access to the dataset, use
     * <code>new Instance(instance)</code>
     *
     * @return the shallow copy
     */
    public Object copy()
    {
        MolInstance result = new MolInstance(this);
        result.m_Dataset = m_Dataset;

        return result;
    }

    /**
     * Returns the description of one instance. If the instance
     * doesn't have access to a dataset, it returns the internal
     * floating-point values. Quotes string
     * values that contain whitespace characters.
     *
     * @return the instance's description as a string
     */
    public String toString()
    {
        StringBuffer text = new StringBuffer();

        // create H depleted molecule
        JOEMol molHdepleted = (JOEMol) mol.clone(false);
        molHdepleted.deleteHydrogens();

        // add SMILES
        text.append(molHdepleted.toString(smiles).trim());
        text.append(',');

        // add title
        String title = molHdepleted.getTitle();
        title = title.replace(',', '_');

        if ((title == null) || (title.trim().length() == 0))
        {
            text.append("unkown");
        }
        else
        {
            text.append(title);
        }

        // add heavy atom coordinates
        JOEAtom atom;
        int charge;
        AtomIterator ait = molHdepleted.atomIterator();

        while (ait.hasNext())
        {
            atom = ait.nextAtom();
            text.append(',');
            text.append(atom.getX());
            text.append(',');
            text.append(atom.getY());
            text.append(',');
            text.append(atom.getZ());
        }

        // add attributes
        text.append(',');

        for (int i = 0; i < m_AttValues.length; i++)
        {
            if (i > 0)
            {
                text.append(",");
            }

            text.append(toString(i));
        }

        //System.out.println(molHdepleted);
        return text.toString();
    }
}
///////////////////////////////////////////////////////////////////////////////
//  END OF FILE.
///////////////////////////////////////////////////////////////////////////////
