///////////////////////////////////////////////////////////////////////////////
//  Filename: $RCSfile: Flat.java,v $
//  Purpose:  Flat file format support.
//  Language: Java
//  Compiler: JDK 1.4
//  Authors:  Joerg K. Wegner
//  Version:  $Revision: 1.11 $
//            $Date: 2004/07/25 20:43:20 $
//            $Author: wegner $
//
//  Copyright (c) Dept. Computer Architecture, University of Tuebingen, Germany
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation version 2 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
///////////////////////////////////////////////////////////////////////////////
package joelib.io.types;

import cformat.PrintfStream;

import wsi.ra.tool.PropertyHolder;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.OutputStream;

import java.util.Properties;
import java.util.Vector;

import org.apache.log4j.Category;

import joelib.data.JOEDataType;
import joelib.data.JOEGenericData;
import joelib.data.JOEPairData;

import joelib.io.IOTypeHolder;
import joelib.io.MoleculeFileType;
import joelib.io.MoleculeIOException;

import joelib.molecule.JOEMol;

import joelib.util.JHM;


/**
 * Flat file format support.
 *
 * @author     wegnerj
 * @license    GPL
 * @cvsversion    $Revision: 1.11 $, $Date: 2004/07/25 20:43:20 $
 */
public class Flat implements MoleculeFileType
{
    //~ Static fields/initializers /////////////////////////////////////////////

    /**
     *  Obtain a suitable logger.
     */
    private static Category logger = Category.getInstance(
            "joelib.io.types.Flat");
    private final static String description = new String(
            "Native descriptor flat file format");
    private final static String[] extensions = new String[]{"flat", "dat", "txt"};

    //~ Instance fields ////////////////////////////////////////////////////////

    public boolean firstLineLoaded = false;
    private LineNumberReader lnr;
    private PrintfStream ps;
    private String delimiter;
    private String inputDelim;
    private String outputDelim;
    private Vector lineStructure;
    private boolean firstLineWritten = false;
    private boolean storeLineInfo = true;
    private int titlePosition;

    //~ Constructors ///////////////////////////////////////////////////////////

    /**
     *  Constructor for the Smiles object
     */
    public Flat()
    {
        lineStructure = new Vector();
    }

    //~ Methods ////////////////////////////////////////////////////////////////

    /**
     *  Sets the storeLineInfo attribute of the Smiles object
     *
     * @param  _flag  The new storeLineInfo value
     */
    public void setStoreLineInfo(boolean _flag)
    {
        storeLineInfo = _flag;
    }

    /**
     *  Gets the storeLineInfo attribute of the Smiles object
     *
     * @return    The storeLineInfo value
     */
    public boolean getStoreLineInfo()
    {
        return storeLineInfo;
    }

    /**
     *  Description of the Method
     *
     * @exception  IOException  Description of the Exception
     */
    public void closeReader() throws IOException
    {
        lnr.close();
    }

    /**
     *  Description of the Method
     *
     * @exception  IOException  Description of the Exception
     */
    public void closeWriter() throws IOException
    {
        ps.close();
    }

    /**
     *  Description of the Method
     *
     * @exception  IOException  Description of the Exception
     */
    public void initParser() throws IOException
    {
        Properties prop = PropertyHolder.instance().getProperties();
        String className = this.getClass().getName();

        // use system properties if available
        String lineStructString = System.getProperty(className +
                ".lineStructure");

        if (lineStructString == null)
        {
            lineStructString = prop.getProperty(className + ".lineStructure",
                    "TITLE");
        }

        //System.out.println("pLINE:"+lineStructString);
        delimiter = prop.getProperty(className + ".lineStructure.delimiter", "|");

        //System.out.println("delimiter: '"+delimiter+"'");
        inputDelim = prop.getProperty(className +
                ".lineStructure.input.delimiter", " \t\n\r");

        //System.out.println("inputDelim:"+inputDelim);
        outputDelim = prop.getProperty(className +
                ".lineStructure.output.delimiter", "\t");

        //System.out.println("outputDelim:"+outputDelim);
        lineStructure.clear();
        JHM.tokenize(lineStructure, lineStructString, delimiter + "\n\r");

        titlePosition = -1;

        for (int i = 0; i < lineStructure.size(); i++)
        {
            if (((String) lineStructure.get(i)).equals("TITLE"))
            {
                titlePosition = i;
            }

            //System.out.println("pSTRUCT("+i+"):"+lineStructure.get(i));
        }
    }

    /**
     *  Description of the Method
     *
     * @param  is               Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    public synchronized void initReader(InputStream is)
        throws IOException
    {
        lnr = new LineNumberReader(new InputStreamReader(is));
        initParser();
        firstLineLoaded = false;
    }

    public synchronized void initReader(InputStream is, boolean _firstLineLoaded)
        throws IOException
    {
        lnr = new LineNumberReader(new InputStreamReader(is));
        initParser();
        firstLineLoaded = _firstLineLoaded;
    }

    /**
     *  Description of the Method
     *
     * @param  os               Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    public void initWriter(OutputStream os) throws IOException
    {
        ps = new PrintfStream(os);
        initParser();
    }

    /**
     *  Description of the Method
     *
     * @return    Description of the Return Value
     */
    public String inputDescription()
    {
        return description;
    }

    /**
     *  Description of the Method
     *
     * @return    Description of the Return Value
     */
    public String[] inputFileExtensions()
    {
        return extensions;
    }

    /**
     *  Description of the Method
     *
     * @return    Description of the Return Value
     */
    public String outputDescription()
    {
        return description;
    }

    /**
     *  Description of the Method
     *
     * @return    Description of the Return Value
     */
    public String[] outputFileExtensions()
    {
        return extensions;
    }

    public String parseFirstLineReadNext(String line) throws IOException
    {
        String nextLine;
        titlePosition = -1;
        lineStructure.clear();
        JHM.tokenize(lineStructure, line, inputDelim + "\n\r");
        firstLineLoaded = true;

        // read first data line
        if ((nextLine = lnr.readLine()) == null)
        {
            logger.error("Flat file contains no data lines.");

            return null;
        }

        StringBuffer sb = new StringBuffer(100);

        for (int i = 0; i < lineStructure.size(); i++)
        {
            sb.append((String) lineStructure.get(i));

            //System.out.println(lineStructure.get(i) + " ");
            if (i < (lineStructure.size() - 1))
            {
                sb.append(delimiter);
            }

            if (((String) lineStructure.get(i)).equals("TITLE"))
            {
                titlePosition = i;
            }
        }

        Properties prop = PropertyHolder.instance().getProperties();
        String className = this.getClass().getName();
        prop.setProperty(className + ".lineStructure", sb.toString());

        logger.info("Set flat file line structure to: " + sb.toString());

        return nextLine;
    }

    /**
     *  Reads an molecule entry as  (unparsed) <tt>String</tt> representation.
     *
     * @param  mol                        the molecule to store the data
     * @return                            <tt>null</tt> if the reader contains no
     *      more relevant data. Otherwise the <tt>String</tt> representation
     *      of the whole molecule entry is returned.
     * @exception  IOException            typical IOException
     */
    public String read() throws IOException
    {
        String line;

        if ((line = lnr.readLine()) == null)
        {
            return null;
        }

        if (!firstLineLoaded)
        {
            line = parseFirstLineReadNext(line);
        }

        return line;
    }

    /**
     *  Description of the Method
     *
     * @param  mol              Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    public boolean read(JOEMol mol) throws IOException, MoleculeIOException
    {
        return read(mol, null);
    }

    /**
     *  Loads an molecule in SMILES format and sets the title. If <tt>title</tt>
     *  is <tt>null</tt> the title line in the molecule file is used.
     *
     * @param  mol              Description of the Parameter
     * @param  title            Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    public synchronized boolean read(JOEMol mol, String title)
        throws IOException, MoleculeIOException
    {
        String line = null;

        if ((line = lnr.readLine()) == null)
        {
            return (false);
        }

        // the first line contains the descriptor data information
        if (!firstLineLoaded)
        {
            line = parseFirstLineReadNext(line);
        }

        Vector data = new Vector();

        // of type String
        JHM.tokenize(data, line, inputDelim);

        if (data.size() < lineStructure.size())
        {
            throw new MoleculeIOException("Line entry \"" + line +
                "\" in line " + lnr.getLineNumber() + " could not be loaded." +
                "Not enough arguments available.");
        }

        String setTitle;

        if (titlePosition == -1)
        {
            if (title == null)
            {
                setTitle = "";
            }
            else
            {
                setTitle = title;
            }
        }
        else
        {
            setTitle = (String) data.get(titlePosition);
        }

        for (int i = 0; i < lineStructure.size(); i++)
        {
            //      System.out.println("");
            if (i != titlePosition)
            {
                //      System.out.println("i:"+lineStructure.get(i));
                if (i < data.size())
                {
                    JOEPairData dp = new JOEPairData();
                    dp.setAttribute((String) lineStructure.get(i));
                    dp.setValue((String) data.get(i));
                    mol.addData(dp);

                    //System.out.println("mol.add: "+lineStructure.get(i)+" "+data.get(i));
                }
                else
                {
                    //        logger.error("Entry \""+(String)lineStructure.get(i)+"\" does not exist in line "+lnr.getLineNumber()+".");
                    throw new IOException("Entry \"" +
                        (String) lineStructure.get(i) +
                        "\" does not exist in line " + lnr.getLineNumber() +
                        ".");
                }
            }
        }

        return (true);
    }

    /**
     *  Description of the Method
     *
     * @return    Description of the Return Value
     */
    public boolean readable()
    {
        return true;
    }

    /**
     *  Description of the Method
     *
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    public boolean skipReaderEntry() throws IOException
    {
        //SMILES line was already readed.
        return true;
    }

    /**
     *  Description of the Method
     *
     * @param  mol              Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    public boolean write(JOEMol mol) throws IOException
    {
        return write(mol, null);
    }

    /**
     *  Description of the Method
     *
     * @param  mol              Description of the Parameter
     * @param  title            Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    public boolean write(JOEMol mol, String title) throws IOException
    {
        // save line structure information
        if (storeLineInfo && !firstLineWritten)
        {
            //ps.print('#');
            for (int i = 0; i < lineStructure.size(); i++)
            {
                if (i != 0)
                {
                    ps.print(delimiter);
                }

                ps.print((String) lineStructure.get(i));
            }

            ps.println();
            firstLineWritten = true;
        }

        // save SMILES line and properties
        for (int i = 0; i < lineStructure.size(); i++)
        {
            if (i != 0)
            {
                ps.print(outputDelim);
            }

            if (i == titlePosition)
            {
                String printTitle = title;

                if (printTitle == null)
                {
                    printTitle = mol.getTitle();
                }

                if (printTitle.trim().equals(""))
                {
                    printTitle = "Undefined";
                }

                ps.print(printTitle);
            }
            else
            {
                JOEGenericData genericData = (JOEGenericData) mol.getData((String) lineStructure.get(
                            i));
                JOEPairData pairData;

                if (genericData == null)
                {
                    logger.warn("Descriptor entry '" + lineStructure.get(i) +
                        "' not found in " + mol.getTitle());
                    logger.warn("Writing NaN");
                    ps.print("NaN");
                }
                else
                {
                    if (genericData.getDataType() == JOEDataType.JOE_PAIR_DATA)
                    {
                        //                ps.print(genericData.getAttribute());
                        pairData = (JOEPairData) genericData;
                        ps.print(pairData.toString(IOTypeHolder.instance()
                                                               .getIOType("FLAT")));
                    }
                }
            }
        }

        ps.println();

        return (true);
    }

    /**
     *  Description of the Method
     *
     * @return    Description of the Return Value
     */
    public boolean writeable()
    {
        return true;
    }
}
///////////////////////////////////////////////////////////////////////////////
//  END OF FILE.
///////////////////////////////////////////////////////////////////////////////
